/*--------------------------------------------------------------------------
	DEFINICJA KLASY UMOZLIWIAJACEJ WYKONYWANIE OBLICZEN NA DUZYCH LICZBACH
	BINARNYCH.
  --------------------------------------------------------------------------

  Reprezentacja liczb binarnych w kodzie uzupelnien do 2 ( U2 )
  pierwszy bit oznacza znak, tzn:
  1 - liczba ujemna
  0 - liczba dodatnia
  -------------------------------------------------------------------------*/

#define  MAX(a,b)       (((a)>(b))?(a):(b))
//przydatne makro - zwraca maximum dwoch liczb

#include <iostream.h>
#include <process.h>
//#include <fstream.h>

typedef enum{FALSE=0,TRUE=1} BOOL; //Typ BOOL przyjmuje False or True

class Binary {
  char *tablica;
  unsigned long rozmiar;
  void BrakPamieci(void);

 public:
  //Binary();					//Konstruktor zwykly bez parametrowy
  Binary(unsigned long rozm=2);   //Konstruktor tworzacy zmienna o danej liczbi bitow
  Binary(const Binary&);//konstr. kopiujacy
  Binary(const char *str);  //Konstruktor tworzacy zmienna z ciagu znakow
  ~Binary();   //destruktor
  friend ostream& operator << (ostream&, Binary&); //pzreciazony operator wysylanie do strumienia
  friend istream& operator >> (istream&, Binary&); //pzreciazony operator wczytywanie ze strumienia
  char operator [] (int);		//operator indexowania
  Binary operator = (Binary&);	//operator przypisania
  Binary operator = (long int);  //operator przypisania
  Binary operator --(); //przedrostkowy dekrementacji
//  Binary operator --(int);//przyrostkowy -- nie zrobiony
  Binary operator ++();     //przyrostkowy inkrementacji
//  Binary operator ++(int);//nie zrobiony
  Binary operator -();  //operator negacji
  friend Binary operator >>(Binary&,int);	//przesuniecie w prawo
  friend Binary operator <<(Binary&,int);  //przesuniecie w lewo
  friend Binary operator + (Binary&, Binary&);
  friend Binary operator - (Binary&, Binary&);
  friend Binary operator * (Binary&, Binary&);
  friend Binary operator / (Binary&, Binary&);

  friend Binary operator ^ (Binary&, int);	//potegowanie
  friend Binary operator ^ (Binary&, Binary&);	//potegowanie
  Binary operator !();                       //silnia

  friend BOOL   operator ==(Binary&, int);     //operatory porownywania
  friend BOOL   operator ==(Binary&, Binary&);
  friend BOOL   operator !=(Binary&, int);
  friend BOOL   operator !=(Binary&, Binary&);
  friend BOOL   operator > (Binary&, int);
  friend BOOL   operator > (Binary&, Binary&);
  friend BOOL   operator >=(Binary&, int);
  friend BOOL   operator >=(Binary&, Binary&);
  friend BOOL   operator < (Binary&, int);
  friend BOOL   operator < (Binary&, Binary&);
  friend BOOL   operator <=(Binary&, int);
  friend BOOL   operator <=(Binary&, Binary&);

  char   sign(void) {return tablica[rozmiar-1];};
		//zwraca znak danej liczby
		//0 dodatnia 1 ujemna
  Binary correct(void);  //deklaracja funkcji do korekcji
  Binary gray(void);  //deklaracja funkcji dla kodu graya
  char* oct(char*);   //deklaracja dla konwersji na kod osemkowy
  char* hex(char*);   //deklaracja dla konwersji na kod szesnastkowy
};

void Binary::BrakPamieci(void){ //gdy zabraknie pamieci na utworzenie nowej zmiennej
	cout << "Brak pamieci na utworzenie zmiennej !!!\n";
	exit(-1);
	};
/*
Binary::Binary() {	//tworzymy 0   Bianary a,c,b
 tablica = new char[2];	//zabezpieczenia liczba nusi miec >=2 bity
 if(tablica==NULL) BrakPamieci();
 rozmiar = 2;
 tablica[0]=tablica[1]=0;
} */

Binary::Binary(unsigned long rozm) {
	if (rozm<=2) {
		//Proba utworzenia zmiennej o liczbie cyfr <= 2
		tablica=new char[2];
		if(tablica==NULL) BrakPamieci();
		rozmiar=2;
		tablica[0]=tablica[1]=0;
		return;
		};
	tablica = new char[rozmiar = rozm];  //tworzy 1tablice dynamiczna
	if(tablica==NULL) BrakPamieci();
	for(int i=0;i<rozm;i++) tablica[i]=0;
}

Binary::Binary(const Binary& liczba) {  //konstruktor kopiujacy
	rozmiar = liczba.rozmiar;
	tablica=new char[rozmiar];          //przydzielanie miejsca
	if(tablica==NULL) BrakPamieci();
	for(int i=0;i<rozmiar;i++)          //kopiowanie do samego siebie bit po bicie
		tablica[i] = liczba.tablica[i];
}

Binary Binary::operator=(Binary &a) {
if (tablica==a.tablica) return *this; //zabezpieczenie przed a=a;
delete tablica;
tablica=new char[a.rozmiar];
if (tablica==NULL) BrakPamieci();
rozmiar=a.rozmiar;        //kopiowanie rozmiaru
for(int i=0;i<rozmiar;i++)	//kopiowanie wszystkich bitow z a do this bit po bicie
	tablica[i]=a.tablica[i];
return *this;
};

Binary Binary::operator=(long int liczba) {  //gdy binarna=int
delete tablica;
rozmiar=sizeof(long int)*8;                  //liczenie liczby bitow
tablica=new char[rozmiar];              //oblicza rozmiar i przydziela pamiec na tablice
if (tablica==NULL) BrakPamieci();
int i;
for(i=0;i<rozmiar;i++) {     //
	tablica[i]=liczba&1;      //dostep do najmniej znaczacego bitu
	liczba>>=1;              // przesowamy liczbe w prawo o jeden to samo co/2
	};
return (*this).correct();
};

Binary::Binary(const char *str) {//tworzenie z ciagu znakow
	//liczba textowo w U2
int i,j,k;
for(i=0,j=0;str[i]!=0;i++)		//zlicza 0 i 1
	if (str[i]=='0'||str[i]=='1') j++;

if (j==0) {//zero cyfr
	tablica=new char[2];
	if (tablica==NULL) BrakPamieci();
	rozmiar=2;
	tablica[0]=tablica[1]=0;
	return;
	};
if (j==1) {//jedna cyfra
	tablica=new char[2];
	if (tablica==NULL) BrakPamieci();
	rozmiar=2;
	for(i=0;str[i]!=0;i++)  //dodaje na poczatek jako najmniej znaczac bit
		if (str[i]=='0'||str[i]=='1')	{
			tablica[0]=(str[i]-'0');
			break;
			};
	tablica[1]=0;        //bardziej znaczacy wstawi 0
	return;
	};
//	liczba cyfr >=2
tablica=new char[j];
if (tablica==NULL) BrakPamieci();
rozmiar=j;
for(i=0,--j;str[i]!=0;i++)
	if (str[i]=='0'||str[i]=='1')
		tablica[j--]=(str[i]-'0');		//najbardziej znaczacy wstawiamy od konca
};

Binary::~Binary() {
delete tablica;
};

Binary Binary::correct(void) {
	//usuwa niepotrzene zera badz jedynki
char szukana=tablica[rozmiar-1]^1; //^1 zmienia 0 na 1 i 1 na 0
int i;
for(i=(rozmiar-2);i>=0;i--) //szukanie 0 lub 1
	if (tablica[i]==szukana) break;
//cout <<endl<< i <<endl;
i=MAX(i,0);               //jesli 00 to obcina do jednego bitu
if ((rozmiar-i-2)>0) {
	//mozna pozbyc sie paru cyfr
	char *new_table=new char[i+2];   //tworzenie nowej tablicy
	if (new_table==NULL) BrakPamieci();
	rozmiar=i+2;
	for(int j=0;j<rozmiar;j++) new_table[j]=tablica[j];       //kopiowanie potrzebnych bitow
	delete tablica;
	tablica=new_table;
	};
return *this;
};

Binary Binary::operator-() {
	//petla od najmniej znaczacego bitu(index 0)
	//szukamy 1 reszte negujemy
int i;
for(i=0;i<rozmiar;i++)
	if (tablica[i]==1) break;
if (i==(rozmiar-1)) { //rozszerzamy wynik o jeden bit
	//np  -(100[=-4])   ->  0100[=4]
	Binary wynik(rozmiar+1);     //nowa zmienna o rozmiarze o 1 wiekszsym
	wynik.tablica[rozmiar-1]=1;
	return wynik;
	};
Binary wynik(*this);
for(++i;i<rozmiar;i++)  //nie trzeba rozszerzac wyniku
	wynik.tablica[i]^=1;
return wynik;
};

Binary Binary::operator --() {
	//petla od najmniej znaczacego bitu(index 0)
	//negujemy bit (0 na 1 ;1 na 0) az napotkamy 1
Binary a="11";//-1 poprzednik
*this=(*this)+a;
return *this;
/*int i;
for(i=0;i<rozmiar;i++) {
	if (tablica[i]==1) break;
	tablica[i]^=1;
	};
if (i==(rozmiar-1)) {//trzeba rozszerzyc wynik o 1 bit
	//np : --(100) ->        1011
	delete tablica;
	tablica=new char[++rozmiar];
	if (tablica==NULL) BrakPamieci();
	for(i=0;i<rozmiar;i++) tablica[i]=1;
	tablica[rozmiar-2]=0;
	//i jest rowne rozmiar wiec nie wykona ponizszego
	};
if (i!=rozmiar)
	//wszystko procz ++(100) i ++(000)
	tablica[i]^=1;
	return *this;
*/
}

Binary Binary::operator ++() {
	//petla od najmniej znaczacego bitu(index 0)
	//negujemy bit (0 na 1 ;1 na 0) az napotkamy 0
Binary a="01";//1 nastepnik
*this=(*this)+a;
return *this;
/*int i;
for(i=0;i<rozmiar;i++) {
	if (tablica[i]==0) break;
	tablica[i]^=1;
	};
if (i==(rozmiar-1)) {//trzeba rozszerzyc wynik o 1 bit
	//np : ++(011)    -> 0100
	delete tablica;
	tablica=new char[++rozmiar];
	if (tablica==NULL) BrakPamieci();
	for(i=0;i<rozmiar;i++) tablica[i]=0;
	tablica[rozmiar-2]=1;
	//i jest rowne rozmiar wiec nie wykona ponizszego
	};
if (i!=rozmiar)
	//wszystko procz ++(111) i ++(011)
	tablica[i]^=1;
return *this;
*/
};

ostream& operator << (ostream& os, Binary& liczba) {
	for(int i=liczba.rozmiar-1;i>=0;--i)
		os << (char)(liczba.tablica[i] + '0');
	return os;
};

istream& operator >> (istream &is,Binary &a){ //wczytywanie i wyswietlanie strumieni
char bufor[1000];
is>>bufor;
a=Binary(bufor);
return is;
};


char Binary::operator [](int indeks) {  //gdy chcemy uzyskac dostep do ktoregos bitu
if (indeks>=rozmiar) return tablica[rozmiar-1];//powielamy bit znaku
if (indeks<0) return 0; //lub 0 gdy indeks <0
return tablica[indeks];  //w przeciwnym wypadku zwraca odpowiedni bit
};

Binary operator + (Binary &a,Binary &b){  //dodawanie
int rozmiar=MAX(a.rozmiar,b.rozmiar); //rozmiar z max liczby a i b
Binary wynik(rozmiar);  //zmienna wynik o odpowiednim rozmiarze
int carry=0,over=0;
	//dzialanie jak w procesorze - symulacja niektorych znacznikow proca
	//carry - przeniesienie
	//over (overflow) - nadmiar
int i;
int tc=0; 	//pomocnicze dla carry
int w;		//tymczasowy wynik
for(i=0;i<rozmiar;i++) {
	w=carry+a[i]+b[i];  //wywolanie przeciazonego operatora indeksowani 2x dla a i b
	tc=w/2;           //czesc calkowita z /2
	over=carry^tc;
	carry=tc;
	wynik.tablica[i]=w%2; //wstawiam do wyniku reszte z dzielenia przez 2
	};
if (over==1) {
	//musimy rozszerzyc wynik o 1 bit
	Binary wyn(rozmiar+1);  //nowa zmienna o rozmiarze +1
	for(i=0;i<rozmiar;i++)
		wyn.tablica[i]=wynik[i]; //kopiejmy wczesniej otrzymany wynik
	wyn.tablica[i]=carry;       //odstawiam najbardziej znaczacy bit
	return wyn;
	};
return wynik;
};

Binary operator - (Binary &a, Binary &b) {  //odejmowanie
return a+(-b);          //korzysta z takze z negacji dla -b
};

Binary operator * (Binary &a, Binary &b) { //mnozenie
int i=0;
Binary wynik(2),bb;
if (b.sign()==1) bb=-b;
else bb=b;
for(i=0;i<bb.rozmiar;i++) {
	//dany bit (i) =0, kontynuuj petle
	if (bb.tablica[i]==0) continue;
	//mnozenie przez 1 odpowiednio przesuniete
	wynik=wynik+(a<<i); //dodaje do wyniku a przesunbiete o ilosc bitow
	};
if (b.sign()==1) return -wynik; //jesli ujemna to zwraca -wynik
return wynik;
};

Binary operator / (Binary &a,Binary &b) { //dzielenie
if (b==0) {  //czy rowna zero dzielnik
	cout<<"Dzielenie przez zero !!!\n";
	return a;
	};
int sign=a.sign()^b.sign(); //obliczanie znaku
Binary reszta=a.correct(),wynik(a.rozmiar); //to co zosatlo oraz to zostalo i ma romzair licznika
Binary mian(b);									  //mianownik taki sam jak b
if (a<0) reszta=-reszta;//pobieranie modolw wykorzystanie operatora przeciazonego "<"
if (b<0) mian=-mian;  //negowanie
int i=a.rozmiar-1; //obliczanie od najbardziej  znaczacego bitu wyniku
for(mian=(mian<<i);i>=0;i--,mian=(mian>>1)) { //zaczynamy od mianownika przyesunietego i poz. w lewo
	if (reszta>=mian) {   //sprawdzanie czy reszta >=mianownik
		wynik.tablica[i]=1;  //wtedy w wyniku 1
		reszta=reszta-mian;    //zmniejszamy reszte o mianownik przesuniety w kazdy kroku o jeden w prawo
		};
	if (reszta==0) break;
	};
if (sign==1) wynik=-wynik; //jesli znak 1 to wynik ujemny
return wynik.correct();   //wycinamy nie potrzebne bity
};

BOOL operator == (Binary &a,Binary &b) { //czy a=b
int rozm=MAX(a.rozmiar,b.rozmiar);  //dla a=5 bit b=3 bit porownoje max=5
for(int i=0;i<rozm;i++)   //sprawdzanie rownosci
	if (a[i]!=b[i]) return FALSE;  //wywolywanie operatora indeksowania przeciaz.
return TRUE;
};

BOOL operator == (Binary &a,int b) {
if (b==0) { //sprawdzanie z intem i dla b=0
	for(int i=0;i<a.rozmiar;i++) if (a[i]==1) return FALSE; //dla kazdego bitu a sprawdz czy ktrorys jest 1
	return TRUE;
	};
Binary temp;
return a==(temp=b); // rozne od zera konwersja na binarna i sprawdzanie
};

BOOL operator != (Binary &a,Binary &b) {  //rozne
if (a==b) return FALSE;
return TRUE;
};

BOOL operator != (Binary &a,int b) {   //to samo co wyzej
if (b==0) {
	for(int i=0;i<a.rozmiar;i++) if (a[i]==1) return TRUE;
	return FALSE;
	};
Binary temp;
return a!=(temp=b);
};


BOOL operator >= (Binary &a,int b) { //>= z int'em
if (b==0) {
	if (a.sign()==1) return FALSE;  //dla - falsz (sprawdzanie)
	return TRUE;
	};
Binary temp;  //gdy b rozne od 0
return a>=(temp=b);
};

BOOL operator >= (Binary &a,Binary &b) { //porownywanie dwoch binarnych >=
return (a-b)>=0; //to samo co a>=b
};

BOOL operator <= (Binary &a,int b) { //<=
if (b==0) {                              //co wyzej dla 0
	if (a.sign()==1) return TRUE;
	return a==0;
	};
Binary temp;
return a<=(temp=b);
};

BOOL operator <= (Binary &a,Binary &b) {   // co wyzej
return (a-b)<=0;
};

BOOL operator > (Binary &a,int b) {  //>
if (b==0) {   // dla 0
	if (a.sign()==1) return FALSE;
	return a!=0;  //czy a rozne od 0
	};
Binary temp;
return a>(temp=b); //co wyzej
};

BOOL operator > (Binary &a,Binary &b) {  //>
return (a-b)>0;
};

BOOL operator < (Binary &a,int b) {
if (b==0) {
	if (a.sign()==1) return TRUE;
	return FALSE;
	};
Binary temp;
return a<(temp=b);
};

BOOL operator < (Binary &a,Binary &b) {
return (a-b)<0;    //obliczy i sprawdzi warunek i przejdzie wyzej
};

Binary operator << (Binary &a,int ile) {//przesuniecie w lewo ile-ile przesowamy bitow
if (ile<0) return a>>(-ile);  //ile <0 przesowamy w prawo o -ile
if (ile==0) return Binary(a); //zwracamy a
Binary wynik(a.rozmiar+ile); //dodaje na koncu 0
for(int i=0;i<a.rozmiar;i++) wynik.tablica[i+ile]=a[i]; //kopiujemy bity i przesowamy o "ile"
return wynik;
};

Binary operator >> (Binary &a,int ile) { //przesuniecie w prawo
if (ile<0) return a<<(-ile);    //war.jak wczesniej
if (ile==0) return Binary(a);
Binary wynik(MAX(2,(int)a.rozmiar-ile)); //musza 2 bity max oblicza rozmiar wyniku
for(int i=0;i<wynik.rozmiar;i++) wynik.tablica[i]=a[i+ile];//bardziej znaczace bity kopiujemy
return wynik;
};

Binary Binary::gray(void){ //kod gray
Binary wynik(*this);  //konstruktor kopiujacy i stwarza wynik na jaki dzialamy
for (int i=0;i<rozmiar-1;i++) //dla wszystkich bitow oprocz najbardzej znaczacego
	 tablica[i]=wynik[i]^wynik[i+1]; //xor
tablica[i]=wynik[i]; //odpowiednie bit z wyniku i zwraca samego siebie
return *this;
};

Binary operator ^ (Binary &a,int n) {
		//potegowanie
if (n==0) return Binary("01"); //do ^0=1
if (n==1) return a;           //do 1^=a
if (a==1) return Binary("01");
if (n<0)  return Binary("00");
return (a^(n/2))*(a^(n/2))*(a^(n%2)); //np a^7=a^3*a^3*a, n%2 resztaz dzielenia przez 2
};

Binary operator ^ (Binary &a,Binary &b) {
		//potegowanie
if (b==0) return Binary("01"); //do ^0=1
if (b==1) return a;           //do 1^=a
if (a==1) return Binary("01");
if (b<0)  return Binary("00");
return (a^(b>>1))*(a^(b>>1))*(a^(b[0])); //np a^7=a^3*a^3*a, n%2 resztaz dzielenia przez 2
};

Binary Binary::operator !() {
Binary jeden("01");
if ((*this)==0) return jeden; //z def !=1
if ((*this)<0)  return (-(!(-*this)));   //jesli mniejsza od zera bire midol obliczam ! i neguje
return (*this)*(!(*this-jeden)); //z def silnia(n)=n*silnia(n-1)
};

char* Binary::oct(char *str){   //wskaznik na wynik
int ile=(rozmiar-1)/3+1,liczba;  //ile cyfr usemkowych do reprezentacji binarnej
str[ile]=0; //str musi sie konczyc zerem
for(int i=0;i<ile;i++) {  //indeksowanie od 0 do ile
	liczba=(*this)[i*3]+(*this)[i*3+1]*2+(*this)[i*3+2]*4; //pobieranie 3 bitow i obliczenie
	str[ile-i-1]=liczba+'0'; //dodajemy 0 i umieszczamy od konca
	};
return str;
};

char* Binary::hex(char *str){
int ile=(rozmiar-1)/4,liczba;//biore paczki 4 bitow
str[ile+1]=0;
for(int i=ile;i>=0;i--) {   //indeksowanie od ile do 0
	liczba=(*this)[i*4]+(*this)[i*4+1]*2+(*this)[i*4+2]*4+(*this)[i*4+3]*8;
	if (liczba<10)	str[ile-i]=liczba+'0'; //jesli jest<10 dodajemu 0
	else str[ile-i]=(liczba-10)+'A';  //jesli >=10 zaczynamy od A
	};
return str;
};

int main() {
 Binary a,b,c;
 char string[1000];
 int i;
 cout<<"\n\n\n\n\n\n\n\n";
 cout<<"Operatory arytmetyczne : \n";
 cout<<"a   = ";cin>>a;
 cout<<"b   = ";cin>>b;
 cout<<"a+b = "<<a+b<<endl;
 cout<<"a-b = "<<a-b<<endl;
 cout<<"a*b = "<<a*b<<endl;
 cout<<"a/b = "<<a/b<<endl;
 cout<<"a^b = "<<(a^b)<<endl;
 cout<<"Metody klasy : \n";
 cout<<"a   = ";cin>>a;
 b=a;
 cout<<"Korekta a = "<<b.correct()<<endl;
 b=a;
 cout<<"Gray    a = "<<b.gray()<<endl;
 b=a;
 cout<<"Poprzednik  a = "<<--b<<endl;
 b=a;
 cout<<"Nastepnik   a = "<<++b<<endl;
 b=a;
 cout<<"Oct    a = "<<b.oct(string)<<endl;
 cout<<"Hex    a = "<<b.hex(string)<<endl;
 cout<<"Silnia a = "<<!b<<endl;
 cout<<"Operatory przesuwania :\n";
 cout<<"a	= ";cin>>a;
 cout<<"i   = ";cin>>i;
 cout<<"a<<i= "<<(a<<i)<<endl;
 cout<<"a>>i= "<<(a>>i)<<endl;
 cout<<"Operatory porownywania :\n";
 cout<<"a    = ";cin>>a;
 cout<<"b    = ";cin>>b;
 cout<<"a==b = "<<(a==b)<<endl;
 cout<<"a!=b = "<<(a!=b)<<endl;
 cout<<"a >b = "<< (a>b)<<endl;
 cout<<"a <b = "<< (a<b)<<endl;
 cout<<"a>=b = "<< (a>=b)<<endl;
 cout<<"a<=b = "<< (a<=b)<<endl;


/*
 int j;
 cin>>j;
 b=j;
 cin>>j;
 a=j;
 cout<<b<<endl;
*/
 return 0;
}